#include <algorithm>
#include <cassert>
#include <cstdlib>
#include <fstream>
#include <ios>
#include <iostream>
#include <numeric>
#include <ostream>
#include <string>

#include <visionaray/math/math.h>
#include <visionaray/aligned_vector.h>

#include "timer.h"

using namespace visionaray;


//-------------------------------------------------------------------------------------------------
// Vertex in untransformed object coordinates
//
// Modifiable in vertex shader. Attributes:
//
// [in/out]
//  pos, normal
//

struct vertex
{
    vertex(vec3& p, vec3& n)
        : pos(p)
        , normal(n)
    {
    }

    // Position
    vec3& pos;

    // Vertex attributes
    vec3& normal;
};


//-------------------------------------------------------------------------------------------------
// Triangle in viewport coordinates
//

struct triangle
{
    vec2 v1;
    vec2 v2;
    vec2 v3;

    // Vertex attributes
    float z1;
    float z2;
    float z3;

    vec3 n1;
    vec3 n2;
    vec3 n3;
};

float area(triangle tri)
{
    vec2 e1 = tri.v2 - tri.v1;
    vec2 e2 = tri.v3 - tri.v1;
    return det2(e2.x, e1.x, e2.y, e1.y) / 2.0f;
}

// bounding rect, integer coords, [min..max)
struct bounds
{
    vec2i min;
    vec2i max;
};

bounds get_bounds(triangle tri)
{
    vec2 minf;
    vec2 maxf;

    minf = tri.v1;
    maxf = floor(tri.v1) + vec2(1,1);

    minf = min(minf, tri.v2);
    maxf = max(maxf, floor(tri.v2) + vec2(1,1));

    minf = min(minf, tri.v3);
    maxf = max(maxf, floor(tri.v3) + vec2(1,1));

    return { vec2i(minf), vec2i(maxf) };
}


//-------------------------------------------------------------------------------------------------
// Fragment helper class
//
// Modifiable in fragment shader. Attributes:
//
// [in]
//  x, y, primitive_id, z, n
//
// [out]
//  color, depth
//

struct fragment
{
    int x;
    int y;

    // Map fragment to id of triangle
    //   (cf. gl_PrimitiveID, APIs guarantee
    //    that fragments in primitive order)
    int primitive_id;

    // Attributes
    float z;
    vec3 n;

    // Output variables
    vec4 color;
    float depth;
};


//-------------------------------------------------------------------------------------------------
// Image class
//
// Store color- and depth buffers etc.
// See typedefs below, use like:
//
// image_rgba32f img1(800, 600);
//  img1(20, 20) = { 1.0f, 0.0f, 0.0f, 0.5f }; // red, 50% transparent
//
// image_rgba8 img2(800, 600);
//  img2(20, 20) = { 0, 255, 0, 127 }; // green, 50% transparent
//
// image_depth32f img3(800, 600);
//  img3(20, 20) = 1.0f; // far away
//

template <typename T>
class image2d
{
public:
    using pixel_type = T;

public:
    image2d()
        : width_(0)
        , height_(0)
        , pixels_(0)
    {
    }

    image2d(int w, int h)
        : width_(w)
        , height_(h)
        , pixels_(w * h)
    {
    }

    image2d(image2d const& rhs)
        : width_(rhs.width_)
        , height_(rhs.height_)
        , pixels_(rhs.width_ * rhs.height_)
    {
        std::copy(rhs.pixels_.begin(), rhs.pixels_.end(), pixels_.begin());
    }

    int width() const
    {
        return width_;
    }

    int height() const
    {
        return height_;
    }

    pixel_type& operator()(int x, int y)
    {
        return pixels_[y * width_ + x];
    }

    pixel_type const& operator()(int x, int y) const
    {
        return pixels_[y * width_ + x];
    }

    pixel_type* data()
    {
        return pixels_.data();
    }

    pixel_type const* data() const
    {
        return pixels_.data();
    }

private:
    int width_;
    int height_;
    aligned_vector<pixel_type> pixels_;

};

using image_rgb8     = image2d<vector<3, unsigned char>>;
using image_rgba8    = image2d<vector<4, unsigned char>>;
using image_rgba32f  = image2d<vector<4, float>>;
using image_depth32f = image2d<float>;


//-------------------------------------------------------------------------------------------------
// perspective() function,
// c.f. man 3 gluPerspective (but gets fovy in rad (not deg!), and *returns* the matrix)
//

mat4 perspective(float fovy, float aspect, float znear, float zfar)
{
    float a = (1/(aspect*tan(fovy/2)));
    float b = (1/(tan(fovy/2)));
    float c = (znear+zfar)/(znear-zfar);
    float d = (2*znear*zfar)/(znear-zfar);

    return mat4(
        a, 0, 0, 0,
        0, b, 0, 0,
        0, 0, c, -1,
        0, 0, d, 0
        );
}


//-------------------------------------------------------------------------------------------------
// Graphics pipeline class, implements the various stages of the graphics pipeline
// and maintains buffers for intermediate outputs (fragments etc.)
//
// This class abstracts a GPU - submit vertices and normals "to VRAM", set pipeline state,
// call the various pipeline stages. Clear buffers in the end to release VRAM ressources.
//

class graphics_pipeline
{
public:

    // Window coordinates reference corner
    enum corner
    {
        BottomLeft, // graphics_pipeline's viewport
        TopLeft     // image formats (like pnm)
    };

    // Rasterization mode
    enum raster_mode
    {
        Points,
        Triangles
    };


    //--- Pipeline state ------------------------------------------------------

    // Enable backface culling
    void enable_backface_culling()
    {
        backface_culling = true;
    }

    // Disable backface culling
    void disable_backface_culling()
    {
        backface_culling = false;
    }

    // Enable depth test
    void enable_depth_test()
    {
        depth_test = true;
    }

    // Disable depth test
    void disable_depth_test()
    {
        depth_test = false;
    }

    // Enable alpha blending
    void enable_alpha_blending()
    {
        alpha_blending = true;
    }

    // Disable alpha blending
    void disable_alpha_blending()
    {
        alpha_blending = false;
    }

    // Set rasterization mode (Points or Triangles)
    void set_raster_mode(raster_mode rm)
    {
        rmode = rm;
    }

    // Viewport transform: set the viewport rectangle: [x,y,width,height]
    void set_viewport(recti vp)
    {
        viewport = vp;
    }

    // Viewport transform: set mapping from nearest and farthest to real values
    void set_depth_range(vec2 dr)
    {
        depth_range = dr;
    }


    //--- Submit buffers ------------------------------------------------------

    // Submit vertex buffer to "VRAM"
    template <typename A>
    void submit_vertex_buffer(std::vector<vec3, A> const& vb)
    {
        vertex_buffer.insert(vertex_buffer.begin(), vb.begin(), vb.end());
    }

    // Submit normal buffer to "VRAM"
    template <typename A>
    void submit_normal_buffer(std::vector<vec3, A> const& nb)
    {
        normal_buffer.insert(normal_buffer.begin(), nb.begin(), nb.end());
    }

    // Submit index buffer to "VRAM"
    template <typename A>
    void submit_index_buffer(std::vector<int, A> const& ib)
    {
        index_buffer.insert(index_buffer.begin(), ib.begin(), ib.end());
    }


    //--- Clear buffers -------------------------------------------------------

    // Clear vertex buffer in "VRAM"
    void clear_vertex_buffer()
    {
        vertex_buffer.clear();
        vertex_buffer.shrink_to_fit();
    }

    // Clear normal buffer in "VRAM"
    void clear_normal_buffer()
    {
        normal_buffer.clear();
        normal_buffer.shrink_to_fit();
    }

    // Clear index buffer in "VRAM"
    void clear_index_buffer()
    {
        index_buffer.clear();
        index_buffer.shrink_to_fit();
    }


    //--- Output --------------------------------------------------------------

    // Utility function: read color buffer to *preallocated* rgb image
    void read_color_buffer(image_rgb8& dest_buffer, corner dest_corner = BottomLeft)
    {
        assert(dest_buffer.width() == color_buffer.width());
        assert(dest_buffer.height() == color_buffer.height());

        // Conversions
        if (dest_corner == BottomLeft)
        {
            for (int y = 0; y < color_buffer.height(); ++y)
            {
                for (int x = 0; x < color_buffer.width(); ++x)
                {
                    dest_buffer(x, y) = color_buffer(x, y).xyz();
                }
            }
        }
        else if (dest_corner == TopLeft)
        {
            for (int y = 0; y < color_buffer.height(); ++y)
            {
                for (int x = 0; x < color_buffer.width(); ++x)
                {
                    dest_buffer(x, color_buffer.height() - y - 1) = color_buffer(x, y).xyz();
                }
            }
        }
    }

    // Utility function: read color buffer to *preallocated* rgba image
    void read_color_buffer(image_rgba8& dest_buffer, corner dest_corner = BottomLeft)
    {
        assert(dest_buffer.width() == color_buffer.width());
        assert(dest_buffer.height() == color_buffer.height());

        // Conversions
        if (dest_corner == BottomLeft)
        {
            // Simple copy
            std::copy(
                color_buffer.data(),
                color_buffer.data() + color_buffer.width() * color_buffer.height(),
                dest_buffer.data()
                );
        }
        else if (dest_corner == TopLeft)
        {
            // Copy scanlines
            for (int y = 0; y < color_buffer.height(); ++y)
            {
                auto src = color_buffer.data() + y * color_buffer.width();
                auto dst = dest_buffer.data() + (color_buffer.height() - y - 1) * color_buffer.width();
                std::copy(src, src + color_buffer.width(), dst);
            }
        }
    }


    //--- Pipeline stages -----------------------------------------------------
    // This API exposes all the programmable and non-programmable graphics
    // pipeline stages. Make sure to call them in order!


    // Vertex stage: pass a program function that expects a *modifiable*
    // vertex as 1st parameter.
    // Pass uniform variables e.g. via lambda capture list
    // Modify v.pos and v.normal.
    template <typename VP>
    void vertex_stage(VP vertex_program)
    {
        timer t;

        for (std::size_t i = 0; i < vertex_buffer.size(); ++i)
        {
            vertex v(vertex_buffer[i], normal_buffer[i]);
            vertex_program(v);
        }

        stats.vertex_stage_timing = t.elapsed();
    }

    // Primitive assembly stage: assemble primitives from vertex stream.
    // This implementation does not perform clipping - make sure
    // that all vertices fit inside the viewport rect!
    void primitive_assembly_stage()
    {
        timer t;

        // Our primitive assembly stage does not perform clipping.
        // At this point we are sure that all triangles fit in
        // the viewport rect and we get #vertices/3 triangles.
        triangle_buffer.resize(vertex_buffer.size() / 3);


        for (std::size_t i = 0; i < index_buffer.size(); i += 3)
        {
            vec3 verts[3] = {
                vertex_buffer[index_buffer[i]],
                vertex_buffer[index_buffer[i+1]],
                vertex_buffer[index_buffer[i+2]]
                };

            // Here we transform normalized device coordinates
            // to viewport coordinates
            for (int j = 0; j < 3; ++j)
            {
                verts[j] = vec3(
                    viewport.w / 2 * verts[j].x + viewport.x + viewport.w / 2,
                    viewport.h / 2 * verts[j].y + viewport.y + viewport.h / 2,
                    (depth_range.y - depth_range.x) / 2 * verts[j].z + (depth_range.y + depth_range.x) / 2
                    );
            }

            triangle tri = {
                    verts[0].xy(),
                    verts[1].xy(),
                    verts[2].xy(),

                    verts[0].z,
                    verts[1].z,
                    verts[2].z,

                    normal_buffer[index_buffer[i]],
                    normal_buffer[index_buffer[i+1]],
                    normal_buffer[index_buffer[i+2]]
                };
            triangle_buffer[i / 3] = tri;
        }

        stats.primitive_assembly_stage_timing = t.elapsed();
    }

    // Scan conversion stage:
    // Points mode:
    //      scan-convert triangle vertices
    // Triangles mode:
    //      scan-convert one triangle after another
    //      and interpolate vertex attributes.
    // Assemble fragment buffer and make sure that API order is restored
    // by storing primitiveID with each fragment
    void scan_conversion_stage()
    {
        timer t;

        for (std::size_t i = 0; i < triangle_buffer.size(); ++i)
        {
            rasterize(triangle_buffer[i], static_cast<int>(i));
        }

        stats.scan_conversion_stage_timing = t.elapsed();
        stats.fragments_in_flight = fragment_buffer.size();
    }

    // Fragment stage: call a fragment shader for each *modifiable* fragment.
    // Pass uniform variables e.g. via lambda capture list
    // Modify f.color and f.depth.
    template <typename FP>
    void fragment_stage(FP fragment_program)
    {
        timer t;

        for (auto& f : fragment_buffer)
        {
            fragment_program(f);
        }

        stats.fragment_stage_timing = t.elapsed();
    }

    // Render output stage: perform per-pixel operations.
    // Clear color and depth buffer.
    // Sort all fragments in API order (APIs like OpenGL guarantee that
    // per-pixel ops are performed in order like in vertex buffer).
    // Perform depth test (TODO)!
    // Support alpha blending (TODO)!
    // Fill color and depth buffers.
    void render_output_stage()
    {
        timer t;

        // Resize output buffers
        color_buffer = image_rgba8(viewport.w, viewport.h);
        depth_buffer = image_depth32f(viewport.w, viewport.h);

        // Clear color to black
        std::fill(
            color_buffer.data(),
            color_buffer.data() + color_buffer.width() * color_buffer.height(),
            vector<4, unsigned char>(U'\0')
            );

        // Clear depth to far
        std::fill(
            depth_buffer.data(),
            depth_buffer.data() + depth_buffer.width() * depth_buffer.height(),
            1.0f
            );

        // Sort fragments by primitive_id to guarantee API order!
        std::sort(
            fragment_buffer.begin(),
            fragment_buffer.end(),
            [](fragment const& a, fragment const& b)
            {
                return a.primitive_id < b.primitive_id;
            }
            );

        // An RGBA 32-bit floating-point buffer to perform alpha and depth compositing in
        image_rgba32f blend_buffer(color_buffer.width(), color_buffer.height());
        std::fill(
            blend_buffer.data(),
            blend_buffer.data() + blend_buffer.width() * blend_buffer.height(),
            vec4(0.0f)
            );

        // TODO: perform depth and alpha test.
        // Input: fragment::color and fragment::depth.
        // Temporarily store colors in blend buffer (perform alpha test w/ FP32 precision!)
        // Perform depth test in depth_buffer.
        // Remember to check pipeline-state (this->depth_test) if
        // depth_test and/or alpha_blending are enabled at all, handle appropriately
        for (auto& f : fragment_buffer)
        {
            blend_buffer(f.x, f.y) = f.color;
            depth_buffer(f.x, f.y) = f.depth;
        }

        // Write blend buffer to color buffer
        for (int y = 0; y < blend_buffer.height(); ++y)
        {
            for (int x = 0; x < blend_buffer.width(); ++x)
            {
                auto src = blend_buffer(x, y);

                // Most basic tone-mapping operator, just clamp
                // everything out of range to [0..1]
                src = clamp(src, vec4(0.0), vec4(1.0));

                // Eliminate alpha
                src.xyz() *= alpha_blending ? src.w : 1.0f;
                src.w = 1.0f;

                // Convert FP32 [0.0..1.0] to bytes
                color_buffer(x, y) = { src.x * 255, src.y * 255, src.z * 255, src.w * 255 };
            }
        }

        stats.render_output_stage_timing = t.elapsed();
    }

    void print_statistics() const
    {
        std::string rmode_str = rmode == Points ? "Points" : "Triangles";

        std::cout << "\n\n";
        std::cout << "#--- Pipeline statistics ------------------------\n";
        std::cout << '\n';
        std::cout << std::boolalpha
                  << "Backface culling: ................. " << backface_culling                      << '\n';
        std::cout << std::boolalpha
                  << "Depth test: ....................... " << depth_test                            << '\n';
        std::cout << std::boolalpha
                  << "Alpha blending: ................... " << alpha_blending                        << '\n';
        std::cout << "Rasterization mode: ............... " << rmode_str                             << '\n';
        std::cout << "Viewport [x,y,w,h]: ............... " <<        viewport.x
                                                            << ',' << viewport.y
                                                            << ',' << viewport.w
                                                            << ',' << viewport.h                     << '\n';
        std::cout << "Vertex stage [sec.]: .............. " << stats.vertex_stage_timing             << '\n';
        std::cout << "Primitive assembly stage [sec.]: .. " << stats.primitive_assembly_stage_timing << '\n';
        std::cout << "Scan conversion stage [sec.]: ..... " << stats.scan_conversion_stage_timing    << '\n';
        std::cout << "Fragment stage [sec.]: ............ " << stats.fragment_stage_timing           << '\n';
        std::cout << "Render output stage [sec.] ........ " << stats.render_output_stage_timing      << '\n';
        std::cout << "Num. fragments in flight: ......... " << stats.fragments_in_flight             << '\n';
        std::cout << "\n\n";
    }

private:

    // Fixed function state
    bool  backface_culling = true;
    bool  depth_test       = true;
    bool  alpha_blending   = false;
    raster_mode rmode = Points;
    recti viewport;
    vec2  depth_range;


    // Storage buffers in "VRAM"
    aligned_vector<vec3> vertex_buffer;
    aligned_vector<vec3> normal_buffer;
    aligned_vector<int>  index_buffer;

    // Stores fragments after fragment stage
    aligned_vector<fragment> fragment_buffer;

    // Stores screen space triangles from primitive assembly
    aligned_vector<triangle> triangle_buffer;


    // Internal color and depth buffer
    image_rgba8    color_buffer;
    image_depth32f depth_buffer;


    // Statistics
    struct
    {
        std::size_t fragments_in_flight        = 0;
        double vertex_stage_timing             = 0.0;
        double primitive_assembly_stage_timing = 0.0;
        double scan_conversion_stage_timing    = 0.0;
        double fragment_stage_timing           = 0.0;
        double render_output_stage_timing      = 0.0;
    } stats;

    // rasterize() function, scan-converts a triangle into the fragment buffer
    void rasterize(triangle tri, int id)
    {
        if (rmode == Points)
        {
            fragment frag1 = { static_cast<int>(tri.v1.x), static_cast<int>(tri.v1.y), id, tri.z1, vec3(), vec4(0.0f), tri.z1 };
            fragment frag2 = { static_cast<int>(tri.v2.x), static_cast<int>(tri.v2.y), id, tri.z2, vec3(), vec4(0.0f), tri.z2 };
            fragment frag3 = { static_cast<int>(tri.v3.x), static_cast<int>(tri.v3.y), id, tri.z3, vec3(), vec4(0.0f), tri.z3 };
            fragment_buffer.push_back(frag1);
            fragment_buffer.push_back(frag2);
            fragment_buffer.push_back(frag3);
        }
        else if (rmode == Triangles)
        {
        }
    }
};


//-------------------------------------------------------------------------------------------------
// Some helper functions to load from and save to files
//

void load_vertices(std::string filename, std::vector<vec3>& vertices)
{
    std::ifstream file(filename);

    std::string line;
    while (std::getline(file, line))
    {
        vec3 v;
        std::istringstream stream(line);
        stream >> v;
        vertices.push_back(v);
    }
}

void load_view_matrix(std::string filename, mat4& view_matrix)
{
    std::ifstream file(filename);

    std::string line;
    std::getline(file, line);
    std::istringstream stream(line);
    stream >> view_matrix;
}

void save_image(std::string filename, image_rgb8 const& img)
{
    std::ofstream file(filename);

    int pitch = img.width() * 3;

    file << "P6\n";
    file << img.width() << ' ' << img.height() << '\n';
    file << 255 << '\n';

    file.write(reinterpret_cast<char const*>(img.data()), pitch * img.height());
}


int main(int argc, char** argv)
{
    // Third arg <output file>: .pnm file (binary format)
    // https://en.wikipedia.org/wiki/Netpbm_format
    //
    // Open with Gimp, Irfanview, etc.

    if (argc < 3)
    {
        std::cerr << "Usage: <program> <vert file> <view file> <output file>\n";
        return EXIT_FAILURE;
    }
   

    //--- Camera parameters -------------------------------

    // Load extr. camera parameters from file
    mat4 view_matrix;
    load_view_matrix(argv[2], view_matrix);

    // The viewport we render into
    recti viewport(0, 0, 1024, 1024);

    // Build perspective transform matrix
    // These values are hardcoded for brevity
    float znear, zfar, fovy, aspect;
    znear = 0.01f;
    zfar = 1.0f;
    fovy = M_PI/4; // 45 degrees
    aspect = static_cast<float>(viewport.w) / viewport.h; // ratio between img's width and height

    mat4 proj_matrix = perspective(fovy, aspect, znear, zfar);

    // When we later transform normals, we do this with
    // the transposed inverse of the view matrix to transform
    // unit vectors to camera space
    mat4 view_matrix_it = transpose(inverse(view_matrix));


    //--- Load vertices from file -------------------------
    std::vector<vec3> vertex_buffer;
    load_vertices(argv[1], vertex_buffer);


    //--- Presort for alpha-compositing -------------------

    // For correct transparency, here's the place we should
    //  1.) construct a temp. copy of all vertices
    //  2.) transform these to eye space
    //  3.) (somehow) maintain a correspondence between the copied vertices
    //      and the original ones
    //  4.) sort the *transformed* vertices by z (eye space: camera is at z=0,
    //      looks in negative z direction)
    //  5.) apply this new ordering *to the original, untransformed* vertices
    // The untransformed vertices are later being transformed in the vertex shader!
    // Don't transform the original vertices here yet, do this in the vertex shader!



    // Build face normals (our pipeline would also support
    // vertex normals, so if you like ;) )
    std::vector<vec3> normal_buffer(vertex_buffer.size());
    for (std::size_t i = 0; i < vertex_buffer.size(); i += 3)
    {
        vec3 e1 = vertex_buffer[i+1] - vertex_buffer[i];
        vec3 e2 = vertex_buffer[i+2] - vertex_buffer[i];
        vec3 n = normalize(cross(e1, e2));
        normal_buffer[i] = n;
        normal_buffer[i+1] = n;
        normal_buffer[i+2] = n;
    }

    // Indices into vertex and normal buffers
    // Here we just use 0,1,2,..
    std::vector<int> index_buffer(vertex_buffer.size());
    std::iota(index_buffer.begin(), index_buffer.end(), 0);


    // Lighting parameters
    vec4 diff(0.8f, 0.5f, 0.4f, 0.5f); // Diffuse material color (RGBA)
    vec3 L(0, 0, 1); // Light direction


    // We need a graphics pipeline object
    graphics_pipeline pipeline;

    // Set up state
    pipeline.enable_backface_culling();
    pipeline.enable_depth_test();
    pipeline.disable_alpha_blending();
    pipeline.set_raster_mode(graphics_pipeline::Points);
//  pipeline.set_raster_mode(graphics_pipeline::Triangles);
    pipeline.set_viewport(viewport);
    pipeline.set_depth_range(vec2(0.0f, 1.0f));

    // Submit vertex data to data queues.
    pipeline.submit_vertex_buffer(vertex_buffer);
    pipeline.submit_normal_buffer(normal_buffer);
    pipeline.submit_index_buffer(index_buffer);

    // Call vertex shader
    pipeline.vertex_stage(
        [view_matrix, proj_matrix, view_matrix_it] // uniform variables
        (vertex& v) // input vertex
        {
            // Convert vertex to ndc-space
            // (modify v.pos)

            // Convert normal to eye-space
            // (modify v.normal)
        }
        );

    // Non-programmable stages
    pipeline.primitive_assembly_stage();
    pipeline.scan_conversion_stage();

    // Call fragment shader
    pipeline.fragment_stage(
        [diff, L] // uniform variables
        (fragment& f) // input fragment
        {
            f.color.xyz() = vec3(1.0f, 1.0f, 1.0f);  // White (RGB)
            f.color.w = 1.0f; // Fully opaque
        }
        );

    // Non-programmable render output stage
    pipeline.render_output_stage();

    // Print some statistics about the pipeline run
    pipeline.print_statistics();

    // Obtain image from pipeline
    image_rgb8 img(viewport.w, viewport.h);
    pipeline.read_color_buffer(
        img,
        graphics_pipeline::TopLeft // Flip from bottom-left to top-left to match .pnm format
        );

    save_image(std::string(argv[3]), img);

    // Clear buffers
    pipeline.clear_vertex_buffer();
    pipeline.clear_normal_buffer();
    pipeline.clear_index_buffer();

    return EXIT_SUCCESS;
}
